require 'pokan'
require 'yagni/hash'
require 'logging'
require 'host'

module Pokan

  require_relative 'pokan-cluster/configuration'
  require_relative 'pokan-cluster/prepared_string'
  require_relative 'pokan-cluster/autorunner'
  require_relative 'pokan-cluster/log_messages'
  require_relative 'pokan-cluster/config_defaults'
  require_relative 'pokan-cluster/version'

  # Cluster module, to be included on all classes that want to implement
  # a custom Gossip server. Including this module enables the developer to
  # customize the configuration file location (and format, as long as she
  # provided a parser for that format). The user can also pass manually
  # the values for the configuration options. However, this is not the
  # recommended way, of course.
  #
  # Example
  #
  #   class MyCustomGossiper
  #     include Pokan::Cluster
  #
  #     config { JSON.parse(CONFIG_PATH) }
  #
  #   end
  #
  # The `config` method expects a hash containing values for network connection,
  # seed address, log information, and others. If you want to see an example of a 
  # valid configuration hash, you can parse the following YAML (recommened config
  # format for pokan-cluster):
  #
  #   pokan:
  #     address: "10.10.10.8" # your IP address
  #
  #     connect:
  #       tcp: 4444
  #       udp: 8787 # ports that will be binded
  #       epoll: true # Linux machine?
  #
  #     gossip:
  #       every: 2 # seconds
  #       share: "cpu,memory,load" # these are the valid options (separated by comma).
  #
  #     seed: "10.10.10.2" # will be contacted on startup
  #
  #     log:
  #       enabled: true # always good!
  #       level: "info"
  #       file: "/path/to/my/log" # default is STDOUT
  #
  # Optionally, you can directly pass a file, using the `config_file` method. However,
  # only the YAML format is currently supported
  #
  #   class MyCustomGossiper
  #     include Pokan::Cluster
  #
  #     config_file = '/path/to/my/config.yml'
  #   end
  module Cluster


    # When a server includes the Pokan::Cluster module, it gains access to its
    # class methods, allowing the user to set configuration and behavior of the
    # server. It is also configured to automatically run when the server class
    # is loaded.
    def self.included(klass)
      klass.extend ClassMethods
      Autorunner.register_server(klass)
    end

    # Automatically starts registered servers.
    at_exit { 
      unless $! || Pokan::Cluster::Autorunner.run?
        Pokan::Cluster::Autorunner.run
      end
    }

    module ClassMethods

      attr_accessor :address, :tcp_port, :udp_port, :epoll, :gossip_interval,
                    :shares, :seed, :logging, :log_level, :log_file

      # Pretty names
      alias_method :epoll?, :epoll
      alias_method :logging?, :logging

      # Loads the server configuration, expecting a valid config hash in a block.
      # This allows you to use whatever configuration format, as long as you can 
      # transform it to a valid pokan hash.
      def config(&block)
        @server || load!(block.call)
        initialize_log
        store_hwinfo
        delegate
      end

      # Convenience method, allowing the user to set the file containing the
      # configuration in the YAML format. In the future, maybe add builtin support
      # for other different formats.
      def config_file(path, options={})
        config { YAML.load(File.read path) }
      end

      # Disables log for the gossip server
      def disable_log
        @logging = false
      end

      # Sets the seed address and redis port, if specified
      #
      #   class Server
      #     include Pokan::Cluster
      #
      #     gossip_with '10.10.10.3', :redis => 6549
      #   end
      #
      # If no Redis port is specified, +6379+ is assumed (as it is the default
      # Redis port).
      def gossip_with(address, options)
        update_autorunner_option(:gossip_with, address)
        update_autorunner_option(:redis, options[:redis] || DEFAULTS[:redis])
      end

      def method_missing(meth, *args, &block)
        @delegator.send(meth, *args, &block)
      end

    private

      # Sets variables according to the hash passed.
      def load!(config_hash)
        config = Configuration.new(config_hash)
        @address = config.get('pokan.address')
        @server = Pokan::Server.new(@address)

        update_ivs(config)
        config_autorunner
      end

      def update_ivs(config)
        @tcp_port         = config.get('pokan.connect.tcp')                 || DEFAULTS[:tcp_port]
        @udp_port         = config.get('pokan.connect.udp')                 || DEFAULTS[:udp_port]
        @epoll            = config.get('pokan.connect.epoll')               || DEFAULTS[:epoll]
        @gossip_interval  = config.get('pokan.gossip.every')                || DEFAULTS[:gossip_interval]
        @seed             = config.get('pokan.seed.address')
        @redis            = config.get('pokan.seed.redis')                  || DEFAULTS[:redis]
        @logging          = config.get('pokan.log.enabled')                 || DEFAULTS[:logging]
        @log_level        = (config.get('pokan.log.level')                   || DEFAULTS[:level]).to_sym
        @log_file         = config.get('pokan.log.file')

        if config.get('pokan.gossip.share').nil?
          @shares = DEFAULTS[:shares]
        else
          @shares = config.get('pokan.gossip.share').split(',')
        end
      end

      def config_autorunner
        options = {}
        options.merge!({ :gossip_with => @seed }) unless @seed.nil?
        options.merge!({ :redis => @redis }) unless @redis.nil?

        Autorunner.options = options
      end

      def update_autorunner_option(key, value)
        opt = Autorunner.options
        opt[key] = value

        Autorunner.options = opt 
      end

      # Adds callbacks for logging on the happening of certain events, such
      # as new keys being defined, changed, gossip messages, etc.
      def initialize_log
        @log_initialized ||= false
        
        if logging? && !@log_initialized
          setup_logging
          add_log_callbacks
        end

        @log_initialized = true
      end

      # requires and sets up the `logging` library.
      def setup_logging
        @logger = Logging.logger['pokan-cluster']
        appenders = [Logging.appenders.stdout]
        appenders << Logging.appenders.file(@log_file) if @log_file

        @logger.add_appenders *appenders
        @logger.level = @log_level
      end

      # adds callbacks on the server instance so that some events are logged
      def add_log_callbacks
        @server.on(:start) do |address, tcp_port, udp_port, epoll, gossip_interval, seed_address, seed_port|
          @logger.debug LOG_MESSAGES[:start][:debug].with(address, tcp_port, udp_port, epoll,
                                                 gossip_interval, seed_address, @log_level.to_s, @log_file).string

          @logger.info LOG_MESSAGES[:start][:info].with(address).string
        end

        @server.before(:sync) do |address, port|
          @logger.debug LOG_MESSAGES[:before_sync][:debug].with(address, port).string
          @logger.info LOG_MESSAGES[:before_sync][:info].string
        end

        @server.after(:sync) do |time_elapsed|
          @logger.debug LOG_MESSAGES[:after_sync][:debug].with(time_elapsed).string
          @logger.info LOG_MESSAGES[:after_sync][:info].string
        end

        @server.on(:new_key) do |key, value|
          @logger.debug LOG_MESSAGES[:new_key][:debug].with(key.to_s, value).string
        end

        @server.before(:gossip) do |address, port, digest_length|
          @logger.debug LOG_MESSAGES[:before_gossip][:debug].with(address, digest_length).string
          @logger.info LOG_MESSAGES[:before_gossip][:info].with(address).string
        end

        @server.after(:gossip) do
          @logger.debug LOG_MESSAGES[:after_gossip][:debug].string
        end

        @server.on(:shutdown) do
          @logger.info LOG_MESSAGES[:shutdown].string
        end


      end

      # Initializes the delegator instance so that unknown methods will
      # be delegated to the server instance, providing a better API for
      # developers who want to customize their server behavior.
      def delegate
        require 'delegate'
        @delegator = DelegateClass(Pokan::Server).new(@server)
      end

      # Stores hardware information (memory, CPU and load average), unless the
      # user specified an empty string in the configuration file (or in a
      # server class).
      def store_hwinfo
        @server.every(@gossip_interlval) { 
          meminfo if @shares.include?('memory')
          cpuinfo if @shares.include?('cpu')
          loadinfo if @shares.include?('load')
        }
      end

      def meminfo
        @server.store(:totalMem, Host::Memory.refresh.total)
        @server.store(:freeMem, Host::Memory.refresh.free)
      end

      def cpuinfo
        @server.store(:nCores, Host::Processor.refresh.cores)
      end

      def loadinfo
        @server.store(:load1, Host::LoadAverage.refresh.on_last(1))
        @server.store(:load5, Host::LoadAverage.refresh.on_last(5))
        @server.store(:load15, Host::LoadAverage.refresh.on_last(15))
        @server.store(:runningProcesses, Host::LoadAverage.refresh.running_processes)
      end

      def prepend(a_string, to_string)
       (a_string.to_s + to_string.to_s).to_sym
      end

    end

  end

end
